#include "WimaMeasurementArea.h"
#include "SnakeTile.h"
#include "snake.h"

#include <boost/units/systems/si.hpp>

const char *WimaMeasurementArea::settingsGroup = "MeasurementArea";
const char *WimaMeasurementArea::tileHeightName = "TileHeight";
const char *WimaMeasurementArea::tileWidthName = "TileWidth";
const char *WimaMeasurementArea::minTileAreaName = "MinTileArea";
const char *WimaMeasurementArea::transectDistanceName = "TransectDistance";
const char *WimaMeasurementArea::minTransectLengthName = "MinTransectLength";
const char *WimaMeasurementArea::showTilesName = "ShowTiles";
const char *WimaMeasurementArea::WimaMeasurementAreaName = "Measurement Area";

WimaMeasurementArea::WimaMeasurementArea(QObject *parent)
    : WimaArea(parent),
      _metaDataMap(FactMetaData::createMapFromJsonFile(
          QStringLiteral(":/json/WimaMeasurementArea.SettingsGroup.json"),
          this /* QObject parent */)),
      _tileHeight(SettingsFact(settingsGroup, _metaDataMap[tileHeightName],
                               this /* QObject parent */)),
      _tileWidth(SettingsFact(settingsGroup, _metaDataMap[tileWidthName],
                              this /* QObject parent */)),
      _minTileArea(SettingsFact(settingsGroup, _metaDataMap[minTileAreaName],
                                this /* QObject parent */)),
      _transectDistance(SettingsFact(settingsGroup,
                                     _metaDataMap[transectDistanceName],
                                     this /* QObject parent */)),
      _minTransectLength(SettingsFact(settingsGroup,
                                      _metaDataMap[minTransectLengthName],
                                      this /* QObject parent */)),
      _showTiles(SettingsFact(settingsGroup, _metaDataMap[showTilesName],
                              this /* QObject parent */)),
      _calculating(false) {
  init();
}

WimaMeasurementArea::WimaMeasurementArea(const WimaMeasurementArea &other,
                                         QObject *parent)
    : WimaArea(other, parent),
      _metaDataMap(FactMetaData::createMapFromJsonFile(
          QStringLiteral(":/json/WimaMeasurementArea.SettingsGroup.json"),
          this /* QObject parent */)),
      _tileHeight(SettingsFact(settingsGroup, _metaDataMap[tileHeightName],
                               this /* QObject parent */)),
      _tileWidth(SettingsFact(settingsGroup, _metaDataMap[tileWidthName],
                              this /* QObject parent */)),
      _minTileArea(SettingsFact(settingsGroup, _metaDataMap[minTileAreaName],
                                this /* QObject parent */)),
      _transectDistance(SettingsFact(settingsGroup,
                                     _metaDataMap[transectDistanceName],
                                     this /* QObject parent */)),
      _minTransectLength(SettingsFact(settingsGroup,
                                      _metaDataMap[minTransectLengthName],
                                      this /* QObject parent */)),
      _showTiles(SettingsFact(settingsGroup, _metaDataMap[showTilesName],
                              this /* QObject parent */)),
      _calculating(false) {
  init();
}

/*!
 * \overload operator=()
 *
 * Calls the inherited operator WimaArea::operator=().
 */
WimaMeasurementArea &WimaMeasurementArea::
operator=(const WimaMeasurementArea &other) {
  WimaArea::operator=(other);

  return *this;
}

WimaMeasurementArea::~WimaMeasurementArea() {
  this->_tiles.clearAndDeleteContents();
}

QString WimaMeasurementArea::mapVisualQML() const {
  return "WimaMeasurementAreaMapVisual.qml";
}

QString WimaMeasurementArea::editorQML() const {
  return "WimaMeasurementAreaEditor.qml";
}

Fact *WimaMeasurementArea::tileHeight() { return &_tileHeight; }

Fact *WimaMeasurementArea::tileWidth() { return &_tileWidth; }

Fact *WimaMeasurementArea::minTileArea() { return &_minTileArea; }

Fact *WimaMeasurementArea::transectDistance() { return &_transectDistance; }

Fact *WimaMeasurementArea::minTransectLength() { return &_minTransectLength; }

Fact *WimaMeasurementArea::showTiles() { return &_showTiles; }

void WimaMeasurementArea::saveToJson(QJsonObject &json) {
  this->WimaArea::saveToJson(json);
  json[tileHeightName] = _tileHeight.rawValue().toDouble();
  json[tileWidthName] = _tileWidth.rawValue().toDouble();
  json[minTileAreaName] = _minTileArea.rawValue().toDouble();
  json[transectDistanceName] = _transectDistance.rawValue().toDouble();
  json[minTransectLengthName] = _minTransectLength.rawValue().toDouble();
  json[showTilesName] = _showTiles.rawValue().toBool();
  json[areaTypeName] = WimaMeasurementAreaName;
}

bool WimaMeasurementArea::loadFromJson(const QJsonObject &json,
                                       QString &errorString) {
  if (this->WimaArea::loadFromJson(json, errorString)) {
    bool retVal = true;

    if (json.contains(tileHeightName) && json[tileHeightName].isDouble()) {
      _tileHeight.setRawValue(json[tileHeightName].toDouble());
    } else {
      errorString.append(tr("Could not load tile height!\n"));
      retVal = false;
    }

    if (json.contains(tileWidthName) && json[tileWidthName].isDouble()) {
      _tileHeight.setRawValue(json[tileWidthName].toDouble());
    } else {
      errorString.append(tr("Could not load tile width!\n"));
      retVal = false;
    }

    if (json.contains(minTileAreaName) && json[minTileAreaName].isDouble()) {
      _tileHeight.setRawValue(json[minTileAreaName].toDouble());
    } else {
      errorString.append(tr("Could not load minimal tile area!\n"));
      retVal = false;
    }

    if (json.contains(transectDistanceName) &&
        json[transectDistanceName].isDouble()) {
      _tileHeight.setRawValue(json[transectDistanceName].toDouble());
    } else {
      errorString.append(tr("Could not load transect distance!\n"));
      retVal = false;
    }

    if (json.contains(minTransectLengthName) &&
        json[minTransectLengthName].isDouble()) {
      _tileHeight.setRawValue(json[minTransectLengthName].toDouble());
    } else {
      errorString.append(tr("Could not load minimal transect length!\n"));
      retVal = false;
    }

    if (json.contains(showTilesName) && json[showTilesName].isBool()) {
      _tileHeight.setRawValue(json[showTilesName].toDouble());
    } else {
      errorString.append(tr("Could not load show tiles !\n"));
      retVal = false;
    }
    return retVal;
  } else {
    return false;
  }
}

void WimaMeasurementArea::doUpdate() {
  using namespace snake;
  using namespace boost::units;
#ifdef SNAKE_SHOW_TIME
  auto start = std::chrono::high_resolution_clock::now();
#endif
  auto polygon = this->coordinateList();
  for (auto &v : polygon) {
    v.setAltitude(0);
  }
  if (polygon.size() > 3) {
    QGeoCoordinate origin = polygon.first();
    BoostPolygon polygonENU;
    areaToEnu(origin, polygon, polygonENU);
    Length height = this->_tileHeight.rawValue().toDouble() * si::meter;
    Length width = this->_tileWidth.rawValue().toDouble() * si::meter;
    Area minArea =
        this->_minTileArea.rawValue().toDouble() * si::meter * si::meter;
    std::vector<BoostPolygon> tilesENU;
    BoundingBox bbox;
    std::string errorString;
    if (snake::tiles(polygonENU, height, width, minArea, tilesENU, bbox,
                     errorString)) {
      this->_tiles.clearAndDeleteContents();
      for (const auto &t : tilesENU) {
        auto geoTile = new SnakeTile(&this->_tiles);
        for (const auto &v : t.outer()) {
          QGeoCoordinate geoVertex;
          fromENU(origin, v, geoVertex);
          geoTile->push_back(geoVertex);
        }
        this->_tiles.append(geoTile);
      }
    }
  }
#ifdef SNAKE_SHOW_TIME
  qDebug() << "WimaMeasurementArea::doUpdate execution time: "
           << std::chrono::duration_cast<std::chrono::milliseconds>(
                  std::chrono::high_resolution_clock::now() - start)
                  .count()
           << " ms";
#endif
}

void WimaMeasurementArea::deferUpdate() {
  if (this->_timer.isActive()) {
    this->_timer.stop();
  }
  this->_timer.start(100);
}

void WimaMeasurementArea::init() {
  this->setObjectName(WimaMeasurementAreaName);
  connect(&this->_tileHeight, &Fact::rawValueChanged, this,
          &WimaMeasurementArea::deferUpdate);
  connect(&this->_tileWidth, &Fact::rawValueChanged, this,
          &WimaMeasurementArea::deferUpdate);
  connect(&this->_minTileArea, &Fact::rawValueChanged, this,
          &WimaMeasurementArea::deferUpdate);
  connect(this, &WimaArea::pathChanged, this,
          &WimaMeasurementArea::deferUpdate);
  this->_timer.setSingleShot(true);
  connect(&this->_timer, &QTimer::timeout, this,
          &WimaMeasurementArea::doUpdate);
}

/*!
 * \class WimaMeasurementArea
 * \brief Class defining the area inside which the actual drone measurements are
 * performed.
 *
 * \sa WimaArea, WimaController, WimaPlaner
 */
